# !/usr/bin/env Python3
# -*- coding: utf-8 -*-
# @Author   : Akili Wang
# @FILE     : feature_select.py
# @Time     : 2022年2月18日   21:15
# @Software : PyCharm
# @Email   : wz19980226@163.com
#   .--,       .--,
#  ( (  \.---./  ) )
#   '.__/o   o\__.'
#      {=  ^  =}
#       >  -  <
#      /       \
#     //       \\                   神兽保佑
#    //|   .   |\\                  代码无bug！
#    "'\       /'"_.-~^`'-.
#       \  _  /--'         `
#     ___)( )(___
#    (((__) (__)))
#    高山仰止，景行行止。虽不能至，心向往之。

import warnings

import numpy as np
from minepy import MINE
from scipy.stats import pearsonr
from sklearn.feature_selection import RFE, SelectFromModel, SelectKBest, VarianceThreshold
from sklearn.linear_model import Lasso, Ridge
from sklearn.model_selection import StratifiedKFold
from sklearn.svm import LinearSVC

from utils.data_process import DataProcessing

warnings.filterwarnings('ignore')  # 禁止出现wanrnings


class FeatureSelection:
    def __init__(self, pro_name='COAD'):
        self.pro_name = pro_name
        self.dp = DataProcessing(pro_name=pro_name)
        self.matrix = self.dp.get_classification_matrix()
        self.label = self.dp.get_raw_label()
        self.clf_data_path = '../dataset/clf_data/'  # 分类部分选择出的特征
        self.select_function = {'TriVote': 'trivote', 'Lasso': 'lasso', 'Ridge': 'ridge', 'Pearson': 'pearson',
                                'Variance': 'variance', 'MIC': 'mic'}

    # 使用TriVote进行特征选择
    def trivote(self, k=70):  # 第一步的C值，目标acc值
        # step1 LinearSVC
        model = LinearSVC(dual=False, random_state=1)
        model.fit(self.matrix, self.label)
        sfm = SelectFromModel(estimator=model, threshold=-np.inf, max_features=int(k * 5), prefit=True)
        matrix = sfm.transform(self.matrix)
        # step2 SVM-RFE
        selector = RFE(estimator=LinearSVC(), n_features_to_select=k * 2)
        matrix = selector.fit_transform(matrix, self.label)
        # step3 LinearSVC
        max_fea = k  # LinearSVC
        t = 0  # 记录循环次数
        matrix_temp = matrix  # 初始化
        while t < 20:  # 控制循环次数，增加特征个数
            t += 1
            # 筛选特征
            lsvc = LinearSVC(dual=False).fit(matrix, self.label)
            model = SelectFromModel(estimator=lsvc, threshold=-np.inf, max_features=max_fea, prefit=True)
            max_fea += 1
            matrix_temp = model.transform(matrix)
            if matrix_temp.shape[1] == 0:  # 如果到这一步的时候没有特征了，则重新开始特征选择
                continue
            kf = StratifiedKFold(n_splits=10, random_state=1, shuffle=True)  # 每次分组都相同
            acc_list = []  # 保存十折中每一折的acc
            for train, val in kf.split(matrix_temp, self.label):
                train_data, train_label = np.array(matrix_temp)[train], np.array(self.label)[train]
                val_data, val_label = np.array(matrix_temp)[val], np.array(self.label)[val]
                clf = LinearSVC(dual=False)
                clf.fit(train_data, train_label)
                predicts = clf.predict(val_data)
                error = 0  # 统计错误个数
                for i in range(len(predicts)):
                    if predicts[i] != val_label[i]:
                        error += 1
                acc = round((len(predicts) - error) / len(predicts), 6)
                acc_list.append(acc)
            mean_acc = np.array(acc_list).mean()  # 十折交叉验证的结果
            if mean_acc >= 1:
                break
        return matrix_temp

    # 使用Lasso进行特征选择
    def lasso(self, k=70):
        model = Lasso().fit(self.matrix, self.label)
        sfm = SelectFromModel(estimator=model, threshold=-np.inf, max_features=k, prefit=True)
        matrix = sfm.transform(self.matrix)
        return matrix

    # 使用Ridge进行特征选择
    def ridge(self, k=70):
        model = Ridge()
        model.fit(self.matrix, self.label)
        sfm = SelectFromModel(estimator=model, threshold=-np.inf, max_features=k, prefit=True)
        matrix = sfm.transform(self.matrix)
        return matrix

    # 使用Pearson进行特征选择
    def pearson(self, k=70):
        # 定义函数钩子
        res = lambda X, Y: np.array(list(map(lambda x: pearsonr(x, Y), X.T))).T[0]
        model = SelectKBest(res, k=k)
        model.fit(self.matrix.astype(float), self.label.astype(float))
        matrix = model.transform(self.matrix)
        return matrix

    # 使用Variance进行特征选择
    def variance(self, k=70):
        model = VarianceThreshold()
        model.fit(self.matrix, self.label)
        mask = np.argsort(-model.variances_)[:k]  # 第一步筛选特征所得下标
        matrix = self.matrix[:, mask]
        return matrix

    # 使用MIC进行特征选择
    def mic(self, k=70):
        def mic(x, y):
            m = MINE()
            m.compute_score(x, y)
            return (m.mic(), 0.1)

        # 定义函数钩子
        res = lambda X, Y: tuple(map(tuple, np.array(list(map(lambda x: mic(x, Y), X.T))).T))
        model = SelectKBest(res, k=k)
        model.fit(self.matrix, self.label)
        matrix = model.transform(self.matrix)
        return matrix

# fs=FeatureSelection('COAD')
# fs.select_features('TriVote')

# pro_list=['LUAD','STAD','LIHC','SARC']
# func_list=['TriVote','Lasso','Ridge','Pearson','Variance','MIC']
# for pro_name in pro_list:
#     fs = FeatureSelection(pro_name)
#     for fs_name in func_list:
#         fs.select_features(fs_name)
