import os

import numpy as np
import pandas as pd
from sklearn.model_selection import StratifiedKFold

from classification.classify import Classification
from classification.evaluate import Evaluateion
from utils.data_process import DataProcessing


class Interface:
    def __init__(self, pro_name="COAD"):
        self.pro_name = pro_name
        self.clf = Classification(pro_name)
        self.dp = DataProcessing(pro_name)
        self.eva = Evaluateion()

    def feature_select(self, save_path, fs_name, k=70):
        sf = getattr(self.clf.fs, self.clf.fs.select_function[fs_name])
        matrix = sf(k)
        with open(save_path, 'w') as file:
            for i in range(matrix.shape[0]):
                for j in range(matrix.shape[1]):
                    file.write(str(matrix[i][j]) + '\t')
                file.write('\n')
        print(self.pro_name + '\t' + fs_name + '\t' + str(k) + ' is already saved! ')

    def data_classify(self, file_name, clf_name='LogisticRegression'):
        matrix = pd.read_table(file_name, sep='\t', header=None, low_memory=False).values
        matrix = matrix[:, :-1]
        label = self.dp.get_raw_label()
        kf = StratifiedKFold(n_splits=10, random_state=1, shuffle=True)
        acc_list = []
        sn_list = []
        sp_list = []
        for train, val in kf.split(matrix, label):
            train_data, train_label = np.array(matrix)[train], np.array(label)[train]
            val_data, val_label = np.array(matrix)[val], np.array(label)[val]
            clf = self.clf.classifier[clf_name]
            clf.fit(train_data, train_label)
            predicts = clf.predict(val_data)
            acc = self.eva.get_acc(val_label, predicts)
            sn = self.eva.get_sn(val_label, predicts)
            sp = self.eva.get_sp(val_label, predicts)
            acc_list.append(acc)
            sn_list.append(sn)
            sp_list.append(sp)
        # print('acc: {}\tsn: {}\tsp: {}\t'.format(np.array(acc_list).mean(), np.array(sn_list).mean(),
        #                                          np.array(sp_list).mean()))
        return np.array(acc_list).mean(), np.array(sn_list).mean(), np.array(sp_list).mean()

    def select_all(self):
        self.clf.find_best_feanum()
        self.clf.get_record()

    def get_record(self):
        pro_list = ['BRCA', 'COAD', 'BLCA', 'UCEC', 'LGG', 'CESC', 'SARC', 'STAD', 'LUAD', 'LIHC']
        file_list = os.listdir(self.clf.clf_data_path)
        file_path = ''
        TriVote_acc = []
        TriVote_sn = []
        TriVote_sp = []
        LR_acc = []
        # table1
        for pro_name in pro_list:
            TriVote_acc.append([])
            TriVote_sn.append([])
            TriVote_sp.append([])
            self.pro_name = pro_name
            self.clf.pro_name = pro_name
            self.dp.pro_name = pro_name
            self.clf.fs.pro_name = pro_name
            self.clf.dp.pro_name = pro_name
            self.clf.fs.dp.pro_name = pro_name
            for file_name in file_list:
                if (self.clf.pro_name in file_name) and ('TriVote' in file_name):
                    file_path = self.clf.clf_data_path + file_name
            for clf_name in self.clf.classifier.keys():
                acc, sn, sp = self.data_classify(file_path, clf_name)
                TriVote_acc[-1].append(acc)
                TriVote_sn[-1].append(sn)
                TriVote_sp[-1].append(sp)
        # table2
        for pro_name in pro_list:
            LR_acc.append([])
            self.pro_name = pro_name
            self.clf.pro_name = pro_name
            self.dp.pro_name = pro_name
            self.clf.fs.pro_name = pro_name
            self.clf.dp.pro_name = pro_name
            self.clf.fs.dp.pro_name = pro_name
            for fs_name in self.clf.fs.select_function.keys():
                for file_name in file_list:
                    if (self.clf.pro_name in file_name) and (fs_name in file_name):
                        file_path = self.clf.clf_data_path + file_name
                acc, _, _ = self.data_classify(file_path, clf_name='LogisticRegression')
                LR_acc[-1].append(acc)
        TriVote_acc, TriVote_sn, TriVote_sp, LR_acc = np.array(TriVote_acc), np.array(TriVote_sn), np.array(
            TriVote_sp), np.array(LR_acc)
        with open(self.clf.clf_record_path + 'TriVote_acc.txt', 'w') as tacc:
            for i in range(TriVote_acc.shape[0]):
                for j in range(TriVote_acc.shape[1]):
                    tacc.write(str(TriVote_acc[i][j]) + '\t')
                tacc.write('\n')
        with open(self.clf.clf_record_path + 'TriVote_sn.txt', 'w') as tsn:
            for i in range(TriVote_sn.shape[0]):
                for j in range(TriVote_sn.shape[1]):
                    tsn.write(str(TriVote_sn[i][j]) + '\t')
                tsn.write('\n')
        with open(self.clf.clf_record_path + 'TriVote_sp.txt', 'w') as tsp:
            for i in range(TriVote_sp.shape[0]):
                for j in range(TriVote_sp.shape[1]):
                    tsp.write(str(TriVote_sp[i][j]) + '\t')
                tsp.write('\n')
        with open(self.clf.clf_record_path + 'LogisticRegression_acc.txt', 'w') as lacc:
            for i in range(LR_acc.shape[0]):
                for j in range(LR_acc.shape[1]):
                    lacc.write(str(LR_acc[i][j]) + '\t')
                lacc.write('\n')


i = Interface('COAD')
i.get_record()
