import warnings

import numpy as np
import torch
import torch.nn as nn

from utils.data_process import DataProcessing

warnings.filterwarnings('ignore')
np.set_printoptions(precision=3)


class VAEWGAN:
    def __init__(self, epochs=50, batch_size=10, lr=1e-2, n1=400, n2=400, gradient=200, num_heads=10, pro_name='COAD'):
        self.dp = DataProcessing(pro_name)
        self.epochs = epochs
        self.batch_size = batch_size
        self.lr = lr
        self.n1, self.n2, self.gradient, self.num_heads = n1, n2, gradient, num_heads
        self.pro_name = pro_name
        self.loss_func = nn.MSELoss()

    def loss_function(self, generated, original, mu, logvar):
        reconstruction_function = nn.MSELoss(size_average=False)
        MSE = reconstruction_function(generated, original)
        KLD_element = mu.pow(2).add_(logvar.exp()).mul_(-1).add_(1).add_(logvar)
        KLD = torch.sum(KLD_element).mul_(-0.5)
        return MSE + KLD

    def pre_train(self, dataLoader, generator, g_optimizer):
        for index, (order, label, data) in enumerate(dataLoader):
            generator.train()
            out1, out21, out22, z, out3 = generator(data)
            loss = self.loss_function(out3, data, out21, out22)
            loss = loss / data.shape[0]
            g_optimizer.zero_grad()
            loss.backward()
            g_optimizer.step()
        return generator, g_optimizer

    def train(self, dataLoader, data_size, generator, g_optimizer, discriminator, d_optimizer):
        train_loss = 0
        hid1, hid2, out, order_list = [], [], [], []
        for index, (order, label, data) in enumerate(dataLoader):
            generator.train()
            discriminator.train()
            real_label = torch.autograd.Variable(torch.ones(len(data)))
            fake_label = torch.autograd.Variable(torch.zeros(len(data)))
            real_out = discriminator(data)
            real_out = real_out.view([len(data)])
            real_loss = -torch.mean(real_out)
            _, _, _, _, fake_data = generator(data)
            fake_out = discriminator(fake_data)
            fake_out = fake_out.view([len(data)])
            fake_loss = torch.mean(fake_out)
            d_loss = fake_loss + real_loss
            d_optimizer.zero_grad()
            d_loss.backward()
            d_optimizer.step()

            out1, out21, out22, z, out3 = generator(data)
            g_loss = self.loss_function(out3, data, out21, out22)
            g_loss = g_loss / data.shape[0]
            train_loss += np.abs(d_loss.item()) * data.shape[0]
            g_optimizer.zero_grad()
            g_loss.backward()
            g_optimizer.step()
            hid1.append(out1)
            hid2.append(z)
            out.append(out3)
            order_list.append(order)
        return train_loss / data_size, generator, g_optimizer, discriminator, d_optimizer, hid1, hid2, out, order_list

    def test(self, dataLoader, data_size, generator, discriminator):
        test_loss = 0
        hid1, hid2, out, order_list = [], [], [], []
        for index, (order, label, data) in enumerate(dataLoader):
            generator.eval()
            real_label = torch.autograd.Variable(torch.ones(len(data)))
            out1, out21, out22, z, out3 = generator(data)
            output = discriminator(out3)
            output = output.view([len(data)])
            d_loss = self.loss_func(output, real_label)
            test_loss += np.abs(d_loss.item()) * data.shape[0]
            hid1.append(out1)
            hid2.append(z)
            out.append(out3)
            order_list.append(order)
        return test_loss / data_size, hid1, hid2, out, order_list
