import gc
import heapq
import time
import warnings

import numpy as np
import pandas as pd
from sklearn.preprocessing import StandardScaler

warnings.filterwarnings('ignore')


class DataPreprocessing:
    def __init__(self):
        self.file_path='../dataset/raw_data/file/'
        self.raw_matrix_path = '../dataset/raw_data/raw_matrix/'
        self.raw_days_path = '../dataset/raw_data/raw_days/'
        self.raw_headers_path = '../dataset/raw_data/raw_headers/'
        self.raw_regression_path = '../dataset/raw_data/raw_regression/'
        self.var_matrix_path = '../dataset/raw_data/var_matrix/'
        self.var_header_path = '../dataset/raw_data/var_header/'
        self.intersection_matrix_path = '../dataset/raw_data/intersection_matrix/'
        self.final_matrix_path = '../dataset/raw_data/final_matrix/'
        self.final_label_path = '../dataset/raw_data/final_label/'
        self.final_days_path = '../dataset/raw_data/final_days/'
        self.raw_matrix5000_path = '../dataset/raw_data/matrix_5000/'

    def get_classify_data(self, pro_name):
        matrix = pd.read_table(self.file_path+pro_name+'.tsv', header=None, sep='\t', low_memory=False).values
        label = pd.read_table(self.final_label_path+pro_name+'.txt', header=None, sep='\t', low_memory=False).values
        sample_name = label[1:, 0]
        name_list = matrix[0, 1:]
        delete_list = []
        for i in range(len(name_list)):
            if name_list[i] not in sample_name:
                delete_list.append(int(i + 1))
        delete_list = np.array(delete_list)
        matrix = np.delete(matrix, delete_list, axis=1)
        matrix = matrix.T
        matrix[0][0] = '000'
        matrix = matrix[np.argsort(matrix[:, 0])]
        temp = matrix[1:, 1:]
        temp = temp.astype(float)
        mask = ~np.isnan(temp).any(axis=0)
        mask = np.insert(mask, 0, 1)
        matrix = matrix[:, mask]
        for i in range(len(matrix[1:, 0])):
            l = matrix[i + 1, 0].split('-')
            matrix[i + 1, 0] = l[1]
        temp = matrix[1:, 1:]
        temp = temp.astype(float)
        variance = np.nanvar(temp, axis=0)
        mask = heapq.nlargest(50000, range(len(variance)), variance.take)
        mask = [int(i + 1) for i in mask]
        mask.insert(0, 0)
        matrix = matrix[:, mask]
        with open(self.raw_matrix5000_path + pro_name+'.txt', 'w+') as file1:
            for i in range(matrix.shape[0]):
                for j in range(matrix.shape[1]):
                    file1.write(str(matrix[i][j]) + '\t')
                file1.write('\n')

    def get_raw_data(self,pro_name):
        matrix = pd.read_table(self.file_path + pro_name + '.tsv', header=None, sep='\t', low_memory=False).values
        label = pd.read_table(self.final_label_path + pro_name + '.txt', header=None, sep='\t', low_memory=False).values
        sample_name = label[1:, 0]
        name_list = matrix[0, 1:]
        delete_list = []
        for i in range(len(name_list)):
            if name_list[i] not in sample_name:
                delete_list.append(int(i + 1))
        delete_list = np.array(delete_list)
        matrix = np.delete(matrix, delete_list, axis=1)
        matrix = matrix.T
        matrix[0][0] = '000'
        matrix = matrix[np.argsort(matrix[:, 0])]
        # matrix=matrix[:, ~np.isnan(matrix).any(axis=0)]
        temp = matrix[1:, 1:]
        temp = temp.astype(float)
        mask = ~np.isnan(temp).any(axis=0)
        mask = np.insert(mask, 0, 1)
        matrix = matrix[:, mask]
        matrix=matrix.T
        for i in range(len(matrix[1:, 0])):
            l = matrix[i + 1, 0].split('-')
            matrix[i + 1, 0] = l[1]
        with open(self.raw_matrix_path+pro_name+'.txt','w+') as file:
            for i in range(matrix.shape[0]):
                for j in range(matrix.shape[1]):
                    file.write(str(matrix[i][j])+'\t')
                file.write('\n')

    def get_raw_header(self, pro_list):
        for pro_name in pro_list:
            print('======' + pro_name + '======')
            data = pd.read_table(self.raw_matrix_path + pro_name, header=None, sep='\t', low_memory=False).values
            header = data[1:, 0]
            with open(self.raw_headers_path + pro_name, 'w+') as file:
                for item in header:
                    file.write(item + '\t')

    def get_regression_data(self,pro_list):
        for pro_name in pro_list:
            print('======' + pro_name + '======')
            data = pd.read_table(self.raw_matrix_path + pro_name, header=None, sep='\t', low_memory=False).values
            days = pd.read_table(self.raw_days_path + pro_name, header=None, sep='\t', low_memory=False).values
            data = data[:, :-1].transpose((1, 0))
            matrix = []
            for item in days[:, 0]:
                for i in range(len(data)):
                    if data[i][0] == item:
                        matrix.append(data[i])
            matrix = np.array(matrix)
            with open(self.raw_regression_path + pro_name, 'w+') as file:
                for i in range(matrix.shape[0]):
                    for j in range(matrix.shape[1]):
                        file.write(str(matrix[i][j]) + '\t')
                    file.write('\n')
            del matrix, file, data
            gc.collect()
            time.sleep(5)

    def get_header_intersection(self,pro_list,year):
        intersection = []
        for pro_name in pro_list:
            print('======' + pro_name + '======')
            temp = pd.read_table(self.raw_headers_path + pro_name, header=None, sep='\t', low_memory=False).values
            temp = temp.squeeze()
            temp = temp[:-1]
            if len(intersection) == 0:
                intersection = temp
            else:
                intersection = list(set(intersection) & (set(temp)))
        with open(self.raw_headers_path + 'intersection'+str(year)+'.txt', 'w+') as f:
            for item in intersection:
                f.write(str(item) + '\t')

    # BLCA=170，COAD=58，UCEC=66，SARC=86，LGG=96，CESC=59，BRCA=71,SARC=70,STAD=139
    def get_matrix_intersection(self,pro_list,year):
        intersection = pd.read_table(self.raw_headers_path + 'intersection'+str(year)+'.txt', header=None, sep='\t',
                                     low_memory=False).values
        intersection = intersection.squeeze()
        intersection = intersection[:-1]
        for pro_name in pro_list:
            print('======' + pro_name + '======')
            matrix = pd.read_table(self.raw_regression_path + pro_name, header=None, sep='\t', low_memory=False).values
            header = pd.read_table(self.raw_headers_path + pro_name, header=None, sep='\t', low_memory=False).values
            matrix = matrix[:, 1:-1]
            matrix = np.transpose(matrix, (1, 0))
            header = header.squeeze()
            header = header[:-1]
            temp = []
            for name in intersection:
                index = np.argwhere(header == name)[0][0]
                temp.append(matrix[index])
            temp = np.transpose(np.array(temp), (1, 0))
            with open(self.intersection_matrix_path + pro_name, 'w+') as f:
                for i in range(temp.shape[0]):
                    for j in range(temp.shape[1]):
                        f.write(str(temp[i][j]) + '\t')
                    f.write('\n')
            del matrix, temp, f
            gc.collect()
            time.sleep(5)

    def get_variance(self, pro_list,year,feature_num=50000):
        pro_dict = {}
        header = pd.read_table(self.raw_headers_path + 'intersection'+str(year)+'.txt', header=None, sep='\t',
                               low_memory=False).values
        header = header.squeeze()
        header = header[1:-1]
        data_all = []
        for pro_name in pro_list:
            print('======' + pro_name + '======')
            data = pd.read_table(self.intersection_matrix_path + pro_name, header=None, sep='\t',
                                 low_memory=False).values
            data = data[:, 1:-1]
            pro_dict[pro_name] = len(data)
            if len(data_all) == 0:
                data_all = data
                print(data_all.shape)
            else:
                data_all = np.vstack((data_all, data))
                print(data_all.shape)
        variance = np.nanvar(data_all, axis=0)
        data_all = np.transpose(data_all, (1, 0))
        z = zip(variance, header, data_all)
        z = sorted(z, reverse=False)
        variance, header, data_all = zip(*z)
        header = np.array(header)
        data_all = np.array(data_all)
        header = header[:feature_num]
        data_all = data_all[:feature_num, :]
        data_all = np.transpose(data_all, (1, 0))
        print('======all======')
        print(data_all.shape)
        with open(self.var_matrix_path + 'all'+str(year)+'.txt', 'w+') as f:
            for i in range(data_all.shape[0]):
                for j in range(data_all.shape[1]):
                    f.write(str(data_all[i][j]) + '\t')
                f.write('\n')
        for pro_name in pro_list:
            print('======' + pro_name + '======')
            data, data_all = data_all[:pro_dict[pro_name], :], data_all[pro_dict[pro_name]:, :]  # 将数组拆分
            print(data.shape)
            with open(self.var_matrix_path + pro_name, 'w+') as data_file:
                for i in range(data.shape[0]):
                    for j in range(data.shape[1]):
                        data_file.write(str(data[i][j]) + '\t')
                    data_file.write('\n')
            del data, data_file
            gc.collect()
            time.sleep(5)
        del data_all, variance
        gc.collect()
        time.sleep(5)
        with open(self.var_header_path + 'intersection'+str(year)+'.txt', 'w+') as header_file:
            for i in range(len(header)):
                header_file.write(str(header[i]) + '\t')
        del header_file, header
        gc.collect()
        time.sleep(5)

    def get_matrix_standardScaler(self,pro_list):
        for pro_name in pro_list:
            matrix = pd.read_table(self.var_matrix_path + pro_name, header=None, sep='\t').values
            matrix = matrix[:, :-1]
            m = StandardScaler()
            matrix = m.fit_transform(matrix)
            with open(self.final_matrix_path + pro_name, 'w+') as f:
                for i in range(matrix.shape[0]):
                    for j in range(matrix.shape[1]):
                        f.write(str(matrix[i][j]) + '\t')
                    f.write('\n')
            del matrix, f, m
            gc.collect()
            time.sleep(2)

    def days_to_fs(self,pro_list,year):
        all_days = []
        for pro_name in pro_list:
            days = pd.read_table(self.raw_days_path + pro_name, header=None, sep='\t', low_memory=False).values
            days = days[1:, 1]
            if len(all_days) == 0:
                all_days = days
                print(all_days.shape)
            else:
                all_days = np.hstack((all_days, days))
                print(all_days.shape)
            days = days.squeeze()
            with open(self.final_days_path + pro_name, 'w+') as f:
                for i in range(len(days)):
                    f.write(str(days[i]) + '\n')
            del f, days
            gc.collect()
        all_days = all_days.squeeze()
        with open(self.final_days_path + 'all'+str(year)+'.txt', 'w+') as f:
            for i in range(len(all_days)):
                f.write(str(all_days[i]) + '\n')

    def preprocess(self,pro_list,year):
        for pro_name in pro_list:
            self.get_classify_data(pro_name)
            self.get_raw_data(pro_name)
        self.get_raw_header(pro_list)
        self.get_regression_data(pro_list)
        self.get_header_intersection(pro_list,year)
        self.get_matrix_intersection(pro_list,year)
        self.get_variance(pro_list,year)
        self.get_matrix_standardScaler(pro_list)
        self.days_to_fs(pro_list,year)


# dp = DataPreprocessing()
# dp.get_raw_header()
# dp.get_regression_data()
# dp.get_header_intersection()
# dp.get_matrix_intersection()
# dp.get_variance()
# dp.get_matrix_standardScaler()
# dp.days_to_fs()
